"""This module provides utilities for managing configurations."""

# Import standard modules
from pathlib import Path
from string import Template
from typing import Optional
from xml.etree.ElementTree import ParseError

from .data import DataError, DataSource, SourceType
from .lang import switch, BatCaveError, BatCaveException, PathName


class ConfigurationError(BatCaveException):
    """Configuration Exceptions.

    Attributes:
        BAD_FORMAT: The configuration file format is invalid.
        BAD_SCHEMA: The configuration schema is not supported.
        CONFIG_NOT_FOUND: The specified configuration file was not found.
    """
    BAD_FORMAT = BatCaveError(1, Template('Bad format for configuration file: $file'))
    BAD_SCHEMA = BatCaveError(2, Template('Invalid schema in configuration file: $file'))
    CONFIG_NOT_FOUND = BatCaveError(3, Template('Unable to find the configuration file: $file'))


class ConfigCollection:
    """This is a container class to hold a collection of configurations read from a file.

    Attributes:
        INCLUDE_CONFIG_TAG: The configuration tag which indicates an include file.
        _CURRENT_CONFIG_SCHEMA: The default DataSource schema to use.
        _MASK_MISSING: The parameter to indicate that child configurations should be ignored.
        _PARAMS_CONFIGURATION: The configuration section which is used as configuration parameters for the collection.
        _PARENT_CONFIGURATION: The parameter to indicate the parent configuration name.
    """
    _CURRENT_CONFIG_SCHEMA = 1
    _MASK_MISSING = 'mask_missing'
    _PARAMS_CONFIGURATION = 'configuration'
    _PARENT_CONFIGURATION = 'parent'

    INCLUDE_CONFIG_TAG = 'include'

    def __init__(self, name: PathName, /, *, create: bool = False, suffix: str = '_config.xml'):
        """
        Args:
            name: The configuration collection name.
            create (optional, default=False): If True the configuration file will be created if it doesn't exist.
            suffix (optional, default=_config.xml): The suffix to add to the name to derive the configuration file name.

        Attributes:
            params: This is the list of configuration parameters read from the configuration section.
            parent: The parent configuration.
            _config_filename: The derived name of the configuration file.
            _configs: This is the collection of configurations.
            _current: This is the value used by the iterator.
            _data_source: A reference to the DataSource instance read from the configuration file.
            _mask_missing: This value is read from the configuration section and
                if True will prevent missing values from the parent from creating values.
            _name: The value of the name argument.

        Raises:
            ConfigurationError.BAD_FORMAT: The the format of the configuration file is not valid.
            ConfigurationError.BAD_SCHEMA: If the schema of the configuration file is not supported.
            ConfigurationError.CONFIG_NOT_FOUND: If the derived configuration file is not found.
        """
        self._name = (path_name := Path(name)).name
        self._config_filename = path_name.parent / (path_name.name + suffix)
        try:
            self._data_source = DataSource(SourceType.xml, self._config_filename, name=self.name, schema=self._CURRENT_CONFIG_SCHEMA, create=create)
        except DataError as err:
            for case in switch(err.code):
                if case(DataError.FILE_OPEN.code):
                    raise ConfigurationError(ConfigurationError.CONFIG_NOT_FOUND, file=self._config_filename) from err
                if case(DataError.BAD_SCHEMA.code):
                    raise ConfigurationError(ConfigurationError.BAD_SCHEMA, file=self._config_filename) from err
                if case():
                    raise
        except ParseError as err:
            raise ConfigurationError(ConfigurationError.BAD_FORMAT, file=self._config_filename) from err

        self.parent: Optional[ConfigCollection] = None
        self._mask_missing = True
        self.params = getattr(self, self._PARAMS_CONFIGURATION) if hasattr(self, self._PARAMS_CONFIGURATION) else None
        if hasattr(self.params, self._PARENT_CONFIGURATION):
            self.parent = ConfigCollection(getattr(self.params, self._PARENT_CONFIGURATION))
        self._mask_missing = hasattr(self.params, self._MASK_MISSING)

        self._configs = [getattr(self, t.name) for t in self._data_source.gettables() if t.name not in (DataSource.INFO_TABLE, self._PARAMS_CONFIGURATION)]
        config_names = [c.name for c in self._configs]
        if self.parent and not self._mask_missing:
            self._configs += [c for c in self.parent if c.name not in config_names]
        self._current = 0

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.close()
        return False

    def __getattr__(self, attr: str):
        if self._data_source.hastable(attr):
            parent_config = getattr(self.parent, attr) if (self.parent and hasattr(self.parent, attr)) else None
            config = Configuration(self._data_source, attr, parent=parent_config)
            if hasattr(config, self.INCLUDE_CONFIG_TAG):
                config = Configuration(self._data_source, attr, parent=parent_config, include=getattr(self, getattr(config, self.INCLUDE_CONFIG_TAG)))
            return config
        if self.parent and not self._mask_missing:
            return getattr(self.parent, attr)
        raise AttributeError(f'Unknown configuration ({attr}) in {self._config_filename}')

    def __iter__(self):
        return self

    def __next__(self):
        if self._current >= len(self._configs):
            raise StopIteration()
        self._current += 1
        return self._configs[self._current - 1]

    name = property(lambda s: s._name, doc='A read-only property which returns the name of the configuration.')

    def add(self, name: str, /) -> str:
        """Add an item to the configuration collection.

        Args:
            name: The name of the item to add.

        Returns:
            The value of the added item.
        """
        self._data_source.addtable(name).addrow()
        self._data_source.commit()
        return getattr(self, name)

    def close(self) -> None:
        """Close the underlying data source."""
        self._data_source.close()


class Configuration:
    """This is a container class to hold an individual configuration in a collection."""

    def __init__(self, config_source: DataSource, name: str, /, *, parent: Optional['Configuration'] = None, include: Optional['Configuration'] = None):
        """
        Args:
            config_source: The configuration source.
            name: The configuration name.
            parent (optional, default=None): If not None, the parent configuration.
            include (optional, default=None): If not None, the configuration to include.

        Attributes:
            _data_source: The value of the config_source argument.
            _data_table: The DataTable holding the configuration values for this configuration.
            _include: The value of the include argument.
            _name: The value of the name argument.
            _parent: The value of the parent argument.
        """
        self._name = name
        self._data_source = config_source
        self._data_table = self._data_source.gettable(self._name)
        self._parent = parent
        self._include = include

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        return False

    def __getattr__(self, attr: str) -> str:
        # First check this configuration to see if it has the requested attribute
        if values := self._data_table.getrows(attr):
            return values[0].getvalue(attr)

        # Next check the include configuration without considering its parents
        if self._include:
            include_parent_value = self._include._parent  # pylint: disable=protected-access
            self._include._parent = None  # pylint: disable=protected-access
            if hasattr(self._include, attr):
                self._include._parent = include_parent_value  # pylint: disable=protected-access
                return getattr(self._include, attr)
            self._include._parent = include_parent_value  # pylint: disable=protected-access

        # Next check the parent configuration considering its includes
        if self._parent and attr != ConfigCollection.INCLUDE_CONFIG_TAG:
            if hasattr(self._parent, attr):
                return getattr(self._parent, attr)

        # Finally check the include configuration considering its parents
        if self._include:
            if hasattr(self._include, attr):
                return getattr(self._include, attr)

        # The attribute wasn't found
        raise AttributeError(f'Unknown parameter ({attr}) for configuration: {self._name}')

    def __setattr__(self, attr: str, value: str) -> None:
        if attr.startswith('_'):
            super().__setattr__(attr, value)
            return
        self._data_table.getrows()[0].setvalue(attr, value)
        self._data_source.commit()

    name = property(lambda s: s._name, doc='A read-only property which returns the name of the configuration.')



"""This module provides utilities for working with versions.

Attributes:
    PYQT_LOADED (bool/str): If not False then it is the string version of the PyQt API.
    VersionStyle (Enum): The version output styles.
"""

# Import standard modules
from enum import Enum
from sys import version as sys_version

# Import third-party modules
try:
    import PyQt5.QtCore as pyqt
except ImportError:
    PYQT_LOADED = ''
else:
    PYQT_LOADED = pyqt.__dict__['PYQT_VERSION_STR']

# Import internal modules
from . import __copyright__, __title__, __version__, __builddate__, __buildname__
from .platarch import Platform

VersionStyle = Enum('VersionStyle', ('full', 'brief', 'oneline', 'aboutbox'))


def get_version_info(style: VersionStyle = VersionStyle.full, /, plattype: str = 'batcave_run', extra_info: str = '') -> str:
    """Get the version information about the currently running application.

    Args:
        style (optional, default=full): The format for the version string to be returned.
        plattype (optional, default='batcave_run'): The platform type for the architecture information in the version string.
        extra_info (optional, default=''): A line to append after the version info but before the copyright.

    Returns:
        Returns the version string.
    """
    plat = getattr(Platform(), plattype)
    if style == VersionStyle.brief:
        return f'{__title__} {__version__}{extra_info} on {plat}'
    if style == VersionStyle.oneline:
        return f'{__title__} {__version__}{extra_info} (Build: {__buildname__}) [{__builddate__}] on {plat}'

    info = ['' if (style == VersionStyle.aboutbox) else f'{__title__} {__version__}',
            'Build: ' + __buildname__,
            'Date: ' + __builddate__,
            'Platform: ' + plat,
            'Python Version: ' + sys_version]
    if PYQT_LOADED:
        info.append('PyQt Version: ' + PYQT_LOADED)
    from . import cms  # pylint: disable=import-outside-toplevel
    if cms.P4_LOADED or cms.GIT_LOADED:
        info.append('CMS Support:')
        if cms.P4_LOADED:
            info.append('    P4Python: ' + cms.P4_LOADED)
        if cms.GIT_LOADED:
            info.append('    GitPython API: ' + cms.GIT_LOADED)
    if extra_info:
        info += [extra_info]
    info.append(__copyright__)
    return '\n'.join(info)

# cSpell:ignore pyqt



"""This module provides utilities for working with load balancers."""

# Import standard modules
from enum import Enum
from string import Template
from time import sleep
from typing import cast, Iterable, Optional, Tuple, Type, Union

# Import third-party modules
from nssrc.com.citrix.netscaler.nitro.resource.config.ssl.sslvserver_sslcertkey_binding import sslvserver_sslcertkey_binding as NetScalerCertificateBinding
from nssrc.com.citrix.netscaler.nitro.resource.config.cache.cachecontentgroup import cachecontentgroup as NetScalerCacheContentGroup
from nssrc.com.citrix.netscaler.nitro.exception.nitro_exception import nitro_exception as NetScalerError
from nssrc.com.citrix.netscaler.nitro.resource.config.lb.lbvserver_responderpolicy_binding import lbvserver_responderpolicy_binding as NetScalerResponderPolicyBinding
from nssrc.com.citrix.netscaler.nitro.resource.config.basic.server import server as NetScalerServer
from nssrc.com.citrix.netscaler.nitro.resource.config.basic.service import service as NetScalerServerService
from nssrc.com.citrix.netscaler.nitro.service.nitro_service import nitro_service as NetScalerService
from nssrc.com.citrix.netscaler.nitro.resource.config.lb.lbvserver import lbvserver as NetScalerVirtualServer
from nssrc.com.citrix.netscaler.nitro.resource.config.lb.lbvserver_service_binding import lbvserver_service_binding as NetScalerVirtualServiceBinding

# Import internal modules
from .lang import BatCaveError, BatCaveException
from .servermgr import get_server_object, Server, ServerType

_STATUS_CHECK_INTERVAL = 2

LbServerSignal = Enum('LbServerSignal', ('enable', 'disable'))
LbType = Enum('LbType', ('netscaler',))
LbVipType = Enum('LbVipType', ('HTTP', 'SSL', 'SSL_OFFLOAD'))


class LoadBalancerError(BatCaveException):
    """Loadbalancer Exceptions.

    Attributes:
        BAD_OBJECT: The object type was not found in the load balancer.
        BAD_SERVER_SIGNAL: The signal is invalid.
    """
    BAD_OBJECT = BatCaveError(1, Template('Requested $type not in load balancer: $name'))
    BAD_SERVER_SIGNAL = BatCaveError(2, Template('Unknown server signal: $signal'))


class LoadBalancer:
    """Class to create a universal abstract interface for a load balancer."""

    def __init__(self, lb_type: LbType, ip_address: str, /, user: str, password: str):
        """
        Args:
            lb_type: The load balancer type.
            ip_address: The hostname or IP address of the load balancer.
            user: The load balancer user for API access.
            password: The load balancer password for API access.

        Attributes:
            ip_address: The value of the ip_address argument.
            type: The value of the lb_type argument.
            _api: The API object for the load balancer

        Raises:
            StateMachineError.BAD_STATUS: if the value of self.status is not in STATE_STATUSES
        """
        self._type = lb_type
        self._ip_address = ip_address
        self._api = NetScalerService(self.ip_address)
        self._api.login(user, password, 3600)

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self._api.logout()
        return False

    ip_address = property(lambda s: s._ip_address, doc='A read-only property which returns the IP address of the load balancer.')
    type = property(lambda s: s._type, doc='A read-only property which returns the type of the load balancer.')

    def add_server(self, server_info: ServerType, /) -> 'LoadBalancerServer':
        """Add the specified server to the load balancer.

        Args:
            server_info: The info about the server to add.

        Returns:
            The result of the server add call.
        """
        server_object = get_server_object(server_info)
        ns_server = NetScalerServer()
        ns_server.name = server_object.hostname
        ns_server.ipaddress = server_object.ip
        NetScalerServer.add(self._api, ns_server)
        return self.get_server(server_info)

    def add_virtual_server(self, server_info: ServerType, /, *, service_type: LbVipType = LbVipType.HTTP,  # pylint: disable=too-many-locals
                           port: Optional[int] = None, servers: Iterable[ServerType] = tuple(),
                           services: Iterable = tuple(), certificates: Iterable = tuple(),
                           responder_policies: Iterable = tuple()) -> Union['LoadBalancerVirtualServer', Tuple['LoadBalancerVirtualServer', 'LoadBalancerVirtualServer']]:
        """Add a virtual server to the load balancer.

        Args:
            server_info: The info about the virtual server to add.
            service_type (optional, default=LbVipType.HTTP): The service type for the virtual server.
            port (optional, default=None): The port for the virtual server, if None the default is based on the service_type.
                LbVipType.HTTP, LbVipType.SSL_OFFLOAD: 80
                LbVipType.SSL: 443
            servers (optional, default=[]): The list of servers to add to the virtual server.
            certificates (optional, default=[]): The list of certificates to add to the virtual server.
            responder_policies (optional, default=[]): The list of responder policies to add to the virtual server.

        Returns:
            The virtual server if there is not offload server, otherwise a tuple of the virtual server and offload server.
        """
        port_value = None
        if port is not None:
            port_value = port
        elif service_type in (LbVipType.HTTP, LbVipType.SSL_OFFLOAD):
            port_value = 80
        elif service_type == LbVipType.SSL:
            port_value = 443
        else:
            port_value = port

        server_objects = [get_server_object(s) for s in (list(servers) if isinstance(servers, (tuple, list)) else [servers])]  # pylint: disable=superfluous-parens
        service_objects = list(services) if isinstance(services, (tuple, list)) else [services]
        certificates = certificates if isinstance(certificates, (tuple, list)) else [certificates]
        responder_policies = responder_policies if isinstance(responder_policies, (tuple, list)) else [responder_policies]

        server_object = get_server_object(server_info)
        ns_virtual_server = NetScalerVirtualServer()
        ns_virtual_server.name = server_object.hostname
        ns_virtual_server.servicetype = 'HTTP' if (service_type == LbVipType.SSL_OFFLOAD) else service_type.name
        ns_virtual_server.ipv46 = server_object.ip
        ns_virtual_server.port = port_value
        NetScalerVirtualServer.add(self._api, ns_virtual_server)
        virtual_server = self.get_virtual_server(server_info)
        offload_server = None

        for server_object in server_objects:
            if not self.has_server(server_object):
                server = self.add_server(server_object)
                service_name = f'svc_{server.name}_{port_value}'
                server.add_service(service_name)
                service_objects.append(service_name)
        for service in service_objects:
            virtual_server.bind_service(service)
        for policy in responder_policies:
            virtual_server.bind_responder_policy(policy)
        if service_type == LbVipType.SSL:
            for cert in certificates:
                virtual_server.bind_certificate(cert)
        elif service_type == LbVipType.SSL_OFFLOAD:
            server_object = get_server_object(server_info)
            offload_server_object = Server(f'{server_object.hostname}_offload', ip=server_object.ip)
            offload_server = self.add_virtual_server(offload_server_object, service_type=LbVipType.SSL, services=service_objects, certificates=certificates)

        return (virtual_server, offload_server) if offload_server else virtual_server  # type: ignore[return-value]

    def flush_cache_content(self, group_name: str, /) -> str:
        """Flush the specified load balancer cache content group.

        Args:
            group_name: The name of the group to flush.

        Returns:
            The result of the flush call.
        """
        return NetScalerCacheContentGroup.flush(self._api, self.get_cache_content_group(group_name))

    def get_cache_content_group(self, group_name: str, /) -> str:
        """Get the specified load balancer cache content group.

        Args:
            group_name: The name of the group to return.

        Returns:
            The requested cache content group.

        Raises:
            LoadBalancerError.BAD_OBJECT: If the requested group could not be found.
        """
        try:
            group_ref = NetScalerCacheContentGroup.get(self._api, group_name)
            group_ref.query = group_ref.host = group_ref.selectorvalue = ''
            return group_ref
        except NetScalerError as err:
            if err.errorcode == 258:
                raise LoadBalancerError(LoadBalancerError.BAD_OBJECT, type='cache content group', name=group_name) from err
            raise

    def get_server(self, server_info: ServerType, /) -> 'LoadBalancerServer':
        """Get the specified server from the load balancer.

        Args:
            server_info: The info about the server to return.

        Returns:
            The requested server.

        Raises:
            LoadBalancerError.BAD_OBJECT: If the requested server could not be found.
        """
        server_hostname = get_server_object(server_info).hostname
        try:
            return LoadBalancerServer(self, NetScalerServer.get(self._api, server_hostname))
        except NetScalerError as err:
            if err.errorcode == 258:
                raise LoadBalancerError(LoadBalancerError.BAD_OBJECT, type='server', name=server_hostname) from err
            raise

    def get_virtual_server(self, server_name: ServerType, /) -> 'LoadBalancerVirtualServer':
        """Get the specified virtual server from the load balancer.

        Args:
            server_name: The name of the virtual server to return.

        Returns:
            The requested server virtual.

        Raises:
            LoadBalancerError.BAD_OBJECT: If the requested virtual server could not be found.
        """
        server_hostname = get_server_object(server_name).hostname
        try:
            return LoadBalancerVirtualServer(self, NetScalerVirtualServer.get(self._api, server_hostname))
        except NetScalerError as err:
            if err.errorcode == 258:
                raise LoadBalancerError(LoadBalancerError.BAD_OBJECT, type='virtual server', name=server_hostname) from err
            raise

    def has_server(self, server_info: ServerType, /) -> bool:
        """Determine if the load balancer server exists.

        Args:
            server_info: The info about the server for which to search.

        Returns:
            True if the requested server exists, False otherwise.
        """
        try:
            self.get_server(server_info)
        except LoadBalancerError as err:
            if err.code == LoadBalancerError.BAD_OBJECT.code:
                return False
            raise
        else:
            return True

    def manage_servers(self, signal: LbServerSignal, servers: Union[ServerType, Iterable[ServerType]], /, *, validate: bool = True) -> None:
        """Perform an action on servers managed by the load balancer.

        Args:
            signal: The action to perform on the managed server.
            servers: The list of servers to manage.
            validate (optional, default=True): Confirm that the request action is complete.

        Returns:
            Nothing.

        Raises:
            LoadBalancerError.BAD_OBJECT: If the requested virtual server could not be found.
        """
        if signal not in (LbServerSignal.enable, LbServerSignal.disable):
            raise LoadBalancerError(LoadBalancerError.BAD_SERVER_SIGNAL, signal=signal.name)

        server_list = servers if isinstance(servers, (tuple, list)) else [cast(ServerType, servers)]
        for server in [get_server_object(s) for s in server_list]:
            getattr(NetScalerServer, signal.name)(self._api, server.hostname)
            while validate and (self.get_server(server).state.lower() != (signal.name + 'd')):
                sleep(_STATUS_CHECK_INTERVAL)


class LoadBalancerObject:
    """Class to create a universal abstract interface for an object in a load balancer."""

    def __init__(self, load_balancer_ref: NetScalerServer, lb_object_ref: Type['LoadBalancerObject'], /):
        """
        Args:
            load_balancer_ref: The load balancer API object reference.
            lb_object_ref: The load balancer object containing this object.

        Attributes:
            _lb_object_ref: The value of the lb_object_ref argument.
            _load_balancer_ref: The value of the load_balancer_ref argument.
        """
        self._load_balancer_ref = load_balancer_ref
        self._lb_object_ref = lb_object_ref

    def __getattr__(self, attr: str) -> str:
        return getattr(self._lb_object_ref, attr)

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        return False


class LoadBalancerServer(LoadBalancerObject):
    """Class to create a universal abstract interface for a load balancer server."""

    def add_service(self, service_name: str, /, *, service_type: str = 'HTTP', port: int = 80) -> 'LoadBalancerService':
        """Add a service to a server in the load balancer.

        Args:
            service_name: The name of the service to add.
            server_type (optional, default='HTTP'): The type of the service to add.
            port (optional, default=80): The port for the service to add.

        Returns:
            The added service.
        """
        ns_service = NetScalerServerService()
        ns_service.servicetype = service_type
        ns_service.name = service_name
        ns_service.servername = self._lb_object_ref.name  # type: ignore[attr-defined]
        ns_service.port = port
        NetScalerServerService.add(self._load_balancer_ref.ns_session, ns_service)
        return self.get_service(service_name)

    def get_service(self, service_name: str, /) -> 'LoadBalancerService':
        """Get the specified service.

        Args:
            service_name: The name of the service to return.

        Returns:
            The requested service.

        Raises:
            LoadBalancerError.BAD_OBJECT: If the requested service could not be found.
        """
        try:
            return LoadBalancerService(self, NetScalerServerService.get(self._load_balancer_ref.ns_session, service_name))
        except NetScalerError as err:
            if err.errorcode == 258:
                raise LoadBalancerError(LoadBalancerError.BAD_OBJECT, type='service', name=service_name) from err
            raise


class LoadBalancerService(LoadBalancerObject):  # pylint: disable=too-few-public-methods
    """Class to create a universal abstract interface for a load balancer service."""


class LoadBalancerVirtualServer(LoadBalancerObject):
    """Class to create a universal abstract interface for a load balancer virtual server."""

    def bind_certificate(self, cert_name: str, /) -> None:
        """Bind a certificate to the virtual server.

        Args:
            cert_name: The name of the certificate to bind.

        Returns:
            Nothing.
       """
        ns_certificate_binding = NetScalerCertificateBinding()
        ns_certificate_binding.vservername = self.name
        ns_certificate_binding.certkeyname = cert_name
        NetScalerCertificateBinding.add(self._load_balancer_ref.ns_session, ns_certificate_binding)

    def bind_responder_policy(self, policy_name: str, /, *, policy_priority: int = 100) -> None:
        """Bind a responder policy to the virtual server.

        Args:
            policy_name: The name of the responder policy to bind.
            policy_priority (optional, default=100): The priority of the policy to bind.

        Returns:
            Nothing.
       """
        ns_policy_binding = NetScalerResponderPolicyBinding()
        ns_policy_binding.name = self.name
        ns_policy_binding.policyname = policy_name
        ns_policy_binding.priority = policy_priority
        NetScalerResponderPolicyBinding.add(self._load_balancer_ref.ns_session, ns_policy_binding)

    def bind_service(self, service: LoadBalancerService, /) -> None:
        """Bind a service to the virtual server.

        Args:
            service: The service to bind.

        Returns:
            Nothing.
       """
        ns_virtual_service_binding = NetScalerVirtualServiceBinding()
        ns_virtual_service_binding.name = self.name
        ns_virtual_service_binding.servicename = service.name if isinstance(service, LoadBalancerService) else service
        NetScalerVirtualServiceBinding.add(self._load_balancer_ref.ns_session, ns_virtual_service_binding)

# cSpell:ignore ipaddress nssrc lbvserver sslcertkey sslvserver vservername



"""This module provides Python language utilities.

Attributes:
    BATCAVE_HOME: The home directory of the module.
    FROZEN (bool): Is this module running in a frozen application. Quick version of sys.frozen
    VALIDATE_PYTHON (bool, default=True): Whether this module should validate the minimum version of Python when loaded.
    WIN32 (bool): Is this module running on a Windows system. Quick version of (sys.platform == 'win32')
"""

# Import standard modules
from dataclasses import dataclass
from os import getenv
from pathlib import Path, PurePath
from string import Template
import sys
from sys import executable, platform, version_info, path as sys_path
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

# Useful contants
FROZEN = getattr(sys, 'frozen', False)
BATCAVE_HOME = Path(executable).parent if FROZEN else Path(sys_path[0])
VALIDATE_PYTHON = True
WIN32 = (platform == 'win32')

CommandResult = Union[str, List[str]]
MessageString = Union[str, Template]
PathName = Union[str, Path, PurePath]


class MsgStr:
    """Class to create a universal abstract interface for message strings.

    This class is only useful when subclassed where the subclass simply defines the _messages.

    Attributes:
        _message: A dictionary of messages provided by subclasses.

    Example:
        class MyMsg(MsgStr):
            _messages = {'Message1': 'This is just a string',
                            'Message2': Template('This is a $what template)}

        where messages are retrieved with
            MyMsg().Message1
            MyMsg(what='this').Message2
    """
    _messages: Dict[str, Union[str, Template]] = dict()

    def __init__(self, instr: Union[str, Template] = '', transform: str = '', **variables):
        """
        Args:
            instr (optional, default=''): The input message string.
            transform (optional, default=None): A string method used to transform the input message string on output.
            variables (optional): A dictionary of variables to pass to the string.Template.substitute method.

        Attributes:
            _str: The value of the instr argument.
            _transform: The value of the transform argument.
            _vars: The value of the variables argument.
        """
        self._str = instr
        self._transform = transform
        self._vars = variables

    def __getattr__(self, attr: str) -> str:
        if attr in list(self._messages.keys()):
            return self._self_to_str(self._messages[attr])
        raise AttributeError(f"'{type(self)}' object has no attribute '{attr}'")

    def __str__(self):
        return self._self_to_str(self._str)

    def _self_to_str(self, _str: MessageString, /) -> str:
        """Convert this message to a string and apply any transforms.

        Returns:
            The string message.
        """
        return_str: str = _str.substitute(self._vars) if isinstance(_str, Template) else _str
        if self._transform:
            return_str = getattr(return_str, self._transform)()
        return return_str


class BatCaveException(Exception, MsgStr):
    """A base class to provide easier Exception management."""
    def __init__(self, errobj: 'BatCaveError', /, **variables):
        """
        Args:
            errobj: The input message string.
            variables (optional): A dictionary of variables to pass to the string.Template.substitute method.

        Attributes:
            vars: The value of the variables argument.
            _errobj: The value of the errobj argument.
        """
        Exception.__init__(self, errobj, variables)
        MsgStr.__init__(self, errobj.msg, **variables)
        self._errobj = errobj
        self.vars = variables

    def __str__(self):
        return MsgStr.__str__(self)

    code = property(lambda s: s._errobj.code, doc='A read-only property which returns the error code from the error object.')


@dataclass(frozen=True)
class BatCaveError:
    """A class to provide an interface for inspecting exceptions.

        Attributes:
            code: A unique error code for this error.
            msg: A user-facing message for this error.
    """
    code: int
    msg: MessageString


class PythonVersionError(BatCaveException):
    """Invalid Python Version Exception.

    Attributes:
        BAD_VERSION: The version of Python is too low.
    """
    BAD_VERSION = BatCaveError(1, Template('Python $needed required but $used used'))


class switch:  # pylint: disable=invalid-name
    """Class to implement a Pythonic switch statement.

    Taken verbatim from: http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/410692

    Title: Readable switch construction without lambdas or dictionaries
    Submitter: Brian Beck (other recipes)
    Last Updated: 2005/04/26
    Version no: 1.7

    Description:
        Python's lack of a 'switch' statement has garnered much discussion and even a PEP.
        The most popular substitute uses dictionaries to map cases to functions, which requires lots of defs or lambdas.
        While the approach shown here may be O(n) for cases, it aims to duplicate C's original 'switch' functionality and
        structure with reasonable accuracy.
    """

    def __init__(self, value: Any, /):
        self.value = value
        self.fall = False
        self.first = True

    def __iter__(self):
        if self.first:
            self.first = False
            yield self.match
        else:
            return

    def match(self, *args) -> bool:
        """Indicate whether or not to enter a case suite."""
        if self.fall or not args:
            return True
        if self.value in args:
            self.fall = True
            return True
        return False


def bool_to_str(expr: Union[bool, str], /) -> str:
    """Converts an expression to a lowercase boolean string value.

    Args:
        expr: The expression to convert.

    Returns:
        'true' if the expression evaluates to True, 'false' otherwise.
    """
    return str(bool(expr)).lower()


def flatten(thing: Iterable[Iterable], /, *, recursive: bool = True) -> Iterable:
    """Flatten an iterable of iterables.

    Args:
        thing: The thing to be flattened.
        recursive (optional, default=True): Whether or not to recursively flatten the item.

    Returns:
        The final single depth item as the same type as thing.
    """
    flattened = False
    result = list()
    for item in thing:
        try:
            result += [i for i in iter(item)]  # pylint: disable=unnecessary-comprehension
            flattened = True
        except TypeError:
            result.append(item)

    if recursive and flattened:
        return flatten(result)

    return type(thing)(result)  # type: ignore[call-arg]


def flatten_string_list(iter_of_string: Iterable[str], /, *, remove_newlines: bool = True) -> str:
    """Flatten an iterable of strings to a single string.

    Args:
        iter_of_string: The list of strings to be flattened to be flattened.
        remove_newlines (optional, default=True): Whether or not to remove newlines from the final list.

    Returns:
        The final string.
    """
    result = ''.join(flatten(iter_of_string))
    if remove_newlines:
        return result.replace('\n', '')
    return result


def is_debug(test_value: Optional[str] = None, /) -> bool:
    """Determine if the BATCAVE_DEBUG environment variable is set.

    Args:
        test_value (optional, default=False): If set, only return true if the value of test_value is in BATCAVE_DEBUG.

    Return:
        True if the OS environment variable BATCAVE_DEBUG is set, False otherwise.
    """
    if not (debug_value := getenv('BATCAVE_DEBUG')):
        return False
    if not test_value:
        return True
    if test_value in debug_value:
        return True
    return False


def str_to_pythonval(the_string: str, /, *, parse_python: bool = False) -> Any:
    """Converts a string to the closest Python object.

    Args:
        the_string: The string to evaluate.
        parse_python (optional, default=False): If the string contains a '~' character, try to convert it to a more complex python object.

    Returns:
        #. If the string represents an integer, return the value as an int.
        #. If the string represents an non-integer number, return the value as a float.
        #. If the string evaluates to 'None' (case-insensitive), return None.
        #. If the string evaluates to 'True' or 'False' (case-insensitive), return True/False.
        #. If parse_python is True and the_string contains '~':
            Split the_string on the first '~' and return the second part as the value of a type specified by the first part.

    Raises:
        ValueError: If the_string is not a string.
    """
    if not isinstance(the_string, str):
        raise ValueError

    if the_string.isdecimal():
        return int(the_string)

    if the_string.isdigit():
        return float(the_string)

    for (test, value) in (('none', None), ('true', True), ('false', False)):
        if the_string.lower() == test:
            return value

    if parse_python and '~' in the_string:
        (data_type, val) = the_string.split('~', 1)
        the_string = eval(f'{data_type}({val})')  # pylint: disable=eval-used

    if parse_python and '~' in the_string:
        (data_type, val) = the_string.split('~', 1)
        the_string = eval(f'{data_type}({val})')  # pylint: disable=eval-used

    return the_string


def validate_python(test_against: Tuple[int, int] = (3, 6), /) -> None:
    """Checks to make sure that a minimum version of Python is used.

    Args:
        test_against (optional, default=(3,7)): The value of Python to check.

    Returns:
        Nothing.

    Raises:
        PythonVersionError.BAD_VERSION: If the version is too low.
    """
    used = version_info[:2]
    needed = test_against if test_against else (3, 6)
    if used != needed:
        raise PythonVersionError(PythonVersionError.BAD_VERSION, used=used, needed=needed)


def xor(value1: Any, value2: Any, /) -> bool:
    """Perform a logical exclusive-or evaluation.

    Args:
        value1, value2: The values on which to perform the operation.

    Returns:
        The logical exclusive-or of the values.
    """
    return bool(value1) ^ bool(value2)



"""Setuptools setup script for BatCave."""

# cSpell:ignore aarch, psutil pywin

from pathlib import Path
from setuptools import find_packages, setup  # Always prefer setuptools over distutils

import batcave
from batcave.fileutil import slurp
from batcave.sysutil import chmod, S_775

# The files need to be writable
chmod(Path.cwd(), S_775, recursive=True)

setup(
    name=batcave.__title__,
    version=batcave.__version__,

    description=batcave.__summary__,
    long_description=''.join(slurp(Path(__file__).parent / 'DOCUMENTATION.md')),
    long_description_content_type='text/markdown',
    keywords='python programming utilities',

    author=batcave.__author__,
    author_email=batcave.__email__,
    license=batcave.__license__,

    url=batcave.__uri__,
    project_urls={
        'Documentation': 'https://batcave.readthedocs.io',
    },

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',

        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.8',

        'Intended Audience :: Developers',
        'Topic :: Software Development',
        'Natural Language :: English',
    ],

    python_requires='~=3.8',
    packages=find_packages(),
    install_requires=['docker ~= 4.4',
                      'GitPython ~= 3.1',
                      'google-cloud',
                      'kubernetes ~= 12.0',
                      'requests ~= 2.25',
                      'pypiwin32 == 223; sys_platform == "win32"',
                      'pywin32 >= 227; sys_platform == "win32"',
                      'pywin32-stubs; sys_platform == "win32"',
                      'WMI ~= 1.5; sys_platform == "win32"',
                      'psutil ~= 5.8; platform_machine not in "arm arm64 armv6l armv7l armv8b armv8l"',
                      'PyQt5 ~= 5.15; platform_machine not in "aarch64 aarch64_be arm arm64 armv6l armv7l armv8b armv8l"'],
    extras_require={
        'dev': ['setuptools', 'twine', 'wheel', 'xmlrunner'],
        'test': []},
    package_data={'batcave': ['py.typed']}
)

# cSpell:ignore armv pypiwin
